package eu.jonahbauer.raytracing;

import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.render.Color;
import eu.jonahbauer.raytracing.render.ImageFormat;
import eu.jonahbauer.raytracing.render.camera.Camera;
import eu.jonahbauer.raytracing.render.camera.SimpleCamera;
import eu.jonahbauer.raytracing.render.canvas.Image;
import eu.jonahbauer.raytracing.render.canvas.LiveCanvas;
import eu.jonahbauer.raytracing.render.material.*;
import eu.jonahbauer.raytracing.render.renderer.SimpleRenderer;
import eu.jonahbauer.raytracing.scene.*;
import eu.jonahbauer.raytracing.scene.hittable2d.Parallelogram;
import eu.jonahbauer.raytracing.scene.hittable3d.Sphere;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class Main {
    public static void main(String[] args) throws IOException {
        var example = getSquares();
        var scene = example.scene();
        var camera = example.camera();

        var renderer = SimpleRenderer.builder()
                .withSamplesPerPixel(500)
                .withMaxDepth(50)
                .withIterative(true)
                .build();

        var image = new LiveCanvas(new Image(camera.getWidth(), camera.getHeight()));
        image.preview();

        long time = System.nanoTime();
        renderer.render(camera, scene, image);
        System.out.printf("rendering finished after %dms", (System.nanoTime() - time) / 1_000_000);

        ImageFormat.PNG.write(image, Path.of("scene-" + System.currentTimeMillis() + ".png"));
    }

    private static @NotNull Example getSpheres() {
        var rng = new Random(1);
        var objects = new ArrayList<Hittable>();
        objects.add(new Sphere(new Vec3(0, -1000, 0), 1000, new LambertianMaterial(new Color(0.5, 0.5, 0.5))));

        for (int a = -11; a < 11; a++) {
            for (int b = -11; b < 11; b++) {
                var center = new Vec3(a + 0.9 * rng.nextDouble(), 0.2, b + 0.9 * rng.nextDouble());
                if (Vec3.distance(center, new Vec3(4, 0.2, 0)) <= 0.9) continue;

                Material material;
                var rnd = rng.nextDouble();
                if (rnd < 0.8) {
                    // diffuse
                    var albedo = Color.multiply(Color.random(rng), Color.random(rng));
                    material = new LambertianMaterial(albedo);
                } else if (rnd < 0.95) {
                    // metal
                    var albedo = Color.random(rng, 0.5, 1.0);
                    var fuzz = rng.nextDouble() * 0.5;
                    material = new MetallicMaterial(albedo, fuzz);
                } else {
                    // glass
                    material = new DielectricMaterial(1.5);
                }

                objects.add(new Sphere(center, 0.2, material));
            }
        }

        objects.add(new Sphere(new Vec3(0, 1, 0), 1.0, new DielectricMaterial(1.5)));
        objects.add(new Sphere(new Vec3(-4, 1, 0), 1.0, new LambertianMaterial(new Color(0.4, 0.2, 0.1))));
        objects.add(new Sphere(new Vec3(4, 1, 0), 1.0, new MetallicMaterial(new Color(0.7, 0.6, 0.5))));

        var camera = SimpleCamera.builder()
                .withImage(1200, 675)
                .withPosition(new Vec3(13, 2, 3))
                .withTarget(new Vec3(0, 0, 0))
                .withFieldOfView(Math.toRadians(20))
                .withFocusDistance(10.0)
                .withBlurAngle(Math.toRadians(0.6))
                .build();

        return new Example(new Scene(getSkyBox(), objects), camera);
    }

    private static @NotNull Example getSquares() {
        return new Example(
                new Scene(
                        getSkyBox(),
                        new Parallelogram(new Vec3(-3, -2, 5), new Vec3(0, 0, -4), new Vec3(0, 4, 0), new LambertianMaterial(new Color(1.0, 0.2, 0.2))),
                        new Parallelogram(new Vec3(-2, -2, 0), new Vec3(4, 0, 0), new Vec3(0, 4, 0), new LambertianMaterial(new Color(0.2, 1.0, 0.2))),
                        new Parallelogram(new Vec3(3, -2, 1), new Vec3(0, 0, 4), new Vec3(0, 4, 0), new LambertianMaterial(new Color(0.2, 0.2, 1.0))),
                        new Parallelogram(new Vec3(-2, 3, 1), new Vec3(4, 0, 0), new Vec3(0, 0, 4), new LambertianMaterial(new Color(1.0, 0.5, 0.0))),
                        new Parallelogram(new Vec3(-2, -3, 5), new Vec3(4, 0, 0), new Vec3(0, 0, -4), new LambertianMaterial(new Color(0.2, 0.8, 0.8)))
                ),
                SimpleCamera.builder()
                        .withImage(400, 400)
                        .withFieldOfView(Math.toRadians(80))
                        .withPosition(new Vec3(0, 0, 9))
                        .withTarget(new Vec3(0, 0, 0))
                        .build()
        );
    }

    private static @NotNull Scene getSimpleScene() {
        return new Scene(
                getSkyBox(),
                new Sphere(new Vec3(0, -100.5, -1.0), 100.0, new LambertianMaterial(new Color(0.8, 0.8, 0.0))),
                new Sphere(new Vec3(0, 0, -1.2), 0.5, new LambertianMaterial(new Color(0.1, 0.2, 0.5))),
                new Sphere(new Vec3(-1.0, 0, -1.2), 0.5, new DielectricMaterial(1.5)),
                new Sphere(new Vec3(-1.0, 0, -1.2), 0.4, new DielectricMaterial(1 / 1.5)),
                new Sphere(new Vec3(1.0, 0, -1.2), 0.5, new MetallicMaterial(new Color(0.8, 0.6, 0.2), 0.0))
        );
    }

    private static @NotNull SkyBox getSkyBox() {
        return SkyBox.gradient(new Color(0.5, 0.7, 1.0), Color.WHITE);
    }

    private record Example(@NotNull Scene scene, @NotNull Camera camera) {}
}